// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2020 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EqpbServer.cc --- implementation of class to return data from
//			     the Filter Wheel Closed data CCF
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 22-11-19: 1.0.0: - original
//
//	AUTHOR:
//		R. Saxton, 2019-11-19
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//

#include "EqpbServer.h"
#include "CalOalUtility.h"
#include "CalServer.h"

#include <algorithm>
#include <cmath>

using namespace CalOalUtility;
using namespace std;

//
//	----------------------( mode parameters )-----------------------------
//

//  Define the EqpbServer class
//
EqpbServer::EqpbServer( 
        EqpbServer::QpbData qpbData,
        CalAtomCollection<EqpbServer> *parent)
        : CalChild<CalAtomCollection<EqpbServer> >(parent), _qpbData(qpbData)
{
}

const EqpbServer::QpbData EqpbServer::qpbData() const
{
    return _qpbData;
}

// Return a column from the QPBdata
/*
const void EqpbServer::getColumn(string &colname)
{
    switch (colname) {

	case "QUAD": { return _qpbData.quad;}
	case "OBSID": { return _qpbData.obsid;}
	case "EXPID": { return _qpbData.expid;}
	case "SUBMODE": { return _qpbData.submode;}
	case "FILTER": { return _qpbData.filter;}
	case "DATE": { return _qpbData.date;}
	case "OBJECT": { return _qpbData.object;}
	case "TOTAL_T": { return _qpbData.total_t;}
	case "CLEAN_T": { return _qpbData.clean_t;}
	case "EXPOSURE": { return _qpbData.exposure;}
	case "BACKSCAL": { return _qpbData.backscal;}
	case "REVOLUT": { return _qpbData.revolut;}
	case "RATE": { return _qpbData.rate;}
	case "RATEERR": { return _qpbData.rateerr;}
	case "RATE1": { return _qpbData.rate1;}
	case "RATE1ERR": { return _qpbData.rate1err;}
	case "RATE2": { return _qpbData.rate2;}
	case "RATE2ERR": { return _qpbData.rate2err;}
	case "RATE3": { return _qpbData.rate3;}
	case "RATE3ERR": { return _qpbData.rate3err;}
	case "RATE4": { return _qpbData.rate4;}
	case "RATE4ERR": { return _qpbData.rate4err;}
	case "HARD1": { return _qpbData.hard1;}
	case "HARD1ERR": { return _qpbData.hard1err;}
	case "HARD2": { return _qpbData.hard2;}
	case "HARD2ERR": { return _qpbData.hard2err;}
	case "HARD3": { return _qpbData.hard3;}
	case "HARD3ERR": { return _qpbData.hard3err;}
	case "SPECP0": { return _qpbData.specp0;}
	case "SPECOOP0": { return _qpbData.specoop0;}
	case "SPECP4": { return _qpbData.specp4;}
	case "SPECOOP4": { return _qpbData.specoop4;}
    
       	default:
            errstr<<"QPB CCF file does not contain column "<<colname
                          <<Sas::error("ColNotKnown");
    }
}
*/

template<>
CalAtomCollection<EqpbServer>::CalAtomCollection(
	                                     const XmmInstrument *instrument,
					     const Ccf *ccf,
					     Xmm::Filter filter,
					     AccuracyLevel accuracy)
{
//	_filter		= filter;
//	_accuracy	= accuracy;

	EqpbServer *null = 0;
	_collection = new vector<EqpbServer *>(1, null);

	CifEntry *e = ccf->entry(instrument, Ccf::QPB);
	DataSet *dataSet = e->open();

      // Open the flare image extension
        Table *table = dataSet->table(e->queryBlockName("QPB"));

      // Define arrays to hold the column data
        //vector<unsigned int> quad;
        //vector<string> obsid, expid, submode, filter, date, object
        //vector<float> total_t, clean_t, exposure, rate, rateerr
        //vector<float> rate1, rate1err, rate2, rate2err, rate3, 
          //            rate3err, rate4, rate4err, hard1, hard1err,
	//	      hard2, hard2err, hard3, hard3err
        //vector<vector<float> > specp0, spec00p0, specp4, spec00p4

        EqpbServer::QpbData qpbdata;
        
      // Read the data
	CalOalUtility::fillVector(table, "QUAD", qpbdata.quad);
	CalOalUtility::fillVector(table, "OBSID", qpbdata.obsid);
	CalOalUtility::fillVector(table, "EXPID", qpbdata.expid);
	CalOalUtility::fillVector(table, "SUBMODE", qpbdata.submode);
	CalOalUtility::fillVector(table, "FILTER", qpbdata.filter);
	CalOalUtility::fillVector(table, "DATE", qpbdata.date);
	CalOalUtility::fillVector(table, "OBJECT", qpbdata.object);
	CalOalUtility::fillVector(table, "TOTAL_T", qpbdata.total_t);
	CalOalUtility::fillVector(table, "CLEAN_T", qpbdata.clean_t);
	CalOalUtility::fillVector(table, "EXPOSURE", qpbdata.exposure);
	CalOalUtility::fillVector(table, "BACKSCAL", qpbdata.backscal);
	CalOalUtility::fillVector(table, "REVOLUT", qpbdata.revolut);
	CalOalUtility::fillVector(table, "RATE", qpbdata.rate);
	CalOalUtility::fillVector(table, "RATEERR", qpbdata.rateerr);
	CalOalUtility::fillVector(table, "RATE1", qpbdata.rate1);
	CalOalUtility::fillVector(table, "RATE1ERR",qpbdata.rate1err);
	CalOalUtility::fillVector(table, "RATE2", qpbdata.rate2);
	CalOalUtility::fillVector(table, "RATE2ERR", qpbdata.rate2err);
	CalOalUtility::fillVector(table, "RATE3", qpbdata.rate3);
	CalOalUtility::fillVector(table, "RATE3ERR", qpbdata.rate3err);
	CalOalUtility::fillVector(table, "RATE4", qpbdata.rate4);
	CalOalUtility::fillVector(table, "RATE4ERR", qpbdata.rate4err);
	CalOalUtility::fillVector(table, "HARD1", qpbdata.hard1);
	CalOalUtility::fillVector(table, "HARD1ERR", qpbdata.hard1err);
	CalOalUtility::fillVector(table, "HARD2", qpbdata.hard2);
	CalOalUtility::fillVector(table, "HARD2ERR", qpbdata.hard2err);
	CalOalUtility::fillVector(table, "HARD3", qpbdata.hard3);
	CalOalUtility::fillVector(table, "HARD3ERR", qpbdata.hard3err);
	//CalOalUtility::fillVector(table, "SPECP0", qpbdata.specp0);
	//CalOalUtility::fillVector(table, "SPECOOP0", qpbdata.specoop0);
	//CalOalUtility::fillVector(table, "SPECP4", qpbdata.specp4);
	//CalOalUtility::fillVector(table, "SPECOOP4", qpbdata.specoop4);

        cout<<"Nrows: "<<table->rows()<<endl;

        ColumnData *cdsp0 =  table->column("SPECP0")->data();
        ColumnData *cdsoop0 =  table->column("SPECOOP0")->data();
        ColumnData *cdsp4 =  table->column("SPECP4")->data();
        ColumnData *cdsoop4 =  table->column("SPECOOP4")->data();
        int32 *p_p0= cdsp0->int32Data();
        int32 *p_oop0= cdsoop0->int32Data();
        int32 *p_p4= cdsp4->int32Data();
        int32 *p_oop4= cdsoop4->int32Data();
        
        //for (unsigned i=0; i<table->rows(); ++i) {
        unsigned int nsiz = 4096;
        vector<int> tmp_p0(nsiz), tmp_oop0(nsiz), 
                    tmp_p4(nsiz), tmp_oop4(nsiz);
        for (unsigned i=0; i<table->rows(); ++i) {
            for (unsigned j=0; j<nsiz; ++j) {
                tmp_p0.push_back(p_p0[i*nsiz+j]);
                tmp_oop0.push_back(p_oop0[i*nsiz+j]);
                tmp_p4.push_back(p_p4[i*nsiz+j]);
                tmp_oop4.push_back(p_oop4[i*nsiz+j]);
            }

            qpbdata.specp0.push_back(tmp_p0);
            qpbdata.specoop0.push_back(tmp_oop0);
            qpbdata.specp4.push_back(tmp_p4);
            qpbdata.specoop4.push_back(tmp_oop4);

            tmp_p0.clear(); tmp_oop0.clear(); 
            tmp_p4.clear(); tmp_oop4.clear();
        }

      // Create the collection - just one
        (*_collection)[0] = new EqpbServer(qpbdata, this);

      // Delete the column pointers
      /*
        delete cdmodeId;
      */
	dataSetServer->close(dataSet);
} // CalAtomCollection<EqpbServer>::CalAtomCollection


/* AI gcc4 porting
 *  too few template-parameter-lists => template<>
 */
template<>
EqpbServer *CalAtomCollection<EqpbServer>::select(
					Ccd,
					Node,
					unsigned) const
{
        //unsigned n = index(modeOffset);

        if (!(*_collection)[0])
           errstr<<"The CCF does not contain Eqpb data "
                          <<Sas::error("NoData");

	return (*_collection)[0];
} // CalAtomCollection<EqpbServer>::select
