// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	HKvalRangesServer.h --- class to retrieve HK validity ranges from CCF
//
//	DESCRIPTION:
//		A class HKvalRangesServer is declared instances of which hold
//		the HK validity ranges for a particular instrument read from
//		CCF::HKParmInt. Each entry is indexed with a keyword which
//		is the name of the HK parameter is applies to.
//
//	CHANGE HISTORY:
//		UL, 1999-10-19: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Tue Oct 19 10:01:55 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/HKvalRangesServer.h,v 1.8 2003/10/03 14:15:41 sasbuild Exp $
//

#ifndef HK_VAL_RANGE_SERVER_H
#define HK_VAL_RANGE_SERVER_H

#include <map>

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"

//
//	---( HKvalRanges )------------------------------------------------------
//
class HKvalRanges : public std::map<std::string, std::string,
									std::less<std::string> > {
public:
	HKvalRanges() {}

	typedef std::map<std::string, std::string, std::less<std::string> >
			MapType;
	typedef MapType::const_iterator MCIter;
	typedef MapType::iterator MIter;

	virtual MapType::const_iterator has(const std::string &key) const;
		//
		//	PURPOSE:
		//		Check whether a validity range for HK parameter <key>
		//		exists; if it does, the returned iterator points to
		//		it, otherwise MapType::end() is given back
		//


	virtual std::string get(const std::string &key) const;
	virtual std::string get(MapType::const_iterator where,
							const std::string &key = "") const;
		//
		//	PURPOSE:
		//		Retrieve the validity range associated with a
		//		HK parameter <key> in the form of a selectlib expression;
		//		the expression can contain the character '@' as a symbolic
		//		reference to the HK parameter name
		//		Please note: attempting to retrieve a non-existing data item
		//		will result in an error condition.
		//


	virtual std::string getExpanded(const std::string &key) const;
	virtual std::string getExpanded(MapType::const_iterator where,
							   const std::string &key = "") const;
		//
		//	PURPOSE:
		//		As get() above but all '@'-references are replaced
		//		with the actual HK parameter names such that the
		//		returned string is a valid selectlib subexpression
		//		


	virtual HKvalRanges reduceWithShortNames(const std::vector<std::string> &
																shortHKnames);
		//
		//	PURPOSE:
		//		Collect all validity ranges with short keys (short ODF
		//		parameter names) into a new HKvalRange object and remove
		//		them from the current one.
		//


	virtual std::string selectionExpression(bool translateLongToShort = true)
	const;
		//
		//	PURPOSE:
		//		The following operations are carried out:
		//			1. Retrieve all available HK validity ranges and convert
		//			   them into selectlib subexpression via getExpanded()
		//			2. all subexpressions are subsequently combined with the
		//			   logical and-operator (&&) to form a single selection
		//			   expression
		//			3. Long HK parameter names are replaced with short ODF
		//			   HK names (controlled by translation table from CCF);
		//			   can be suppressed by setting <translateLongToShort>
		//			   to false
		//

	virtual ~HKvalRanges() {}
		//
		//	PURPOSE:
		//		Destrory HKvalRanges object constructed with above
		//		constructor(s).
		//


	const MapType &translationMap(const MapType &map)
		{ _translationMap = map; return _translationMap; }
	const MapType &translationMap()
		{ return _translationMap; }

protected:
	MapType	_translationMap;

private:
	std::string expand(const std::string &expr, const std::string &key) const;
};


//
//	---( HKvalRangesServer )--------------------------------------------------
//
class HKvalRangesServer :
			public CalAtom,
			public HKvalRanges,
			public virtual CalChild<CalAtomCollection<HKvalRangesServer> > {
public:
	HKvalRangesServer(CalAtomCollection<HKvalRangesServer> *parent = 0)
		: CalChild<CalAtomCollection<HKvalRangesServer> >(parent) {}
		//
		//	PURPOSE:
		//		Instantiate a HKvalRangesServer object
		//


	static unsigned id() { return Ccf::HKParmInt; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(std::ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual HKvalRanges &valRanges();
		//
		//	PURPOSE:
		//		Give access to HKvalRanges object
		//


	virtual ~HKvalRangesServer() {}
		//
		//	PURPOSE:
		//		Destrory HKvalRangesServer object constructed with above
		//		constructor(s).
		//
};

#endif
