// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	PsfDataServer.cc --- implementation of PsfDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the PsfDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-09-05: 1.0.0: - first version
//		UL, 2000-09-25: 1.1.0: - bug fix: _xOffset/_yOffset were not
//								 initialized
//		UL, 2000-10-17: 1.2.0: - added yet another flavor of
//									XmmPsf::enboxedEnergy()
//								 that simply integrates Psf in given window
//							   - all enboxed/encircledEnergy() variants now
//								 make use of spatial origin set via
//								 atOrigin()
//		UL, 2000-10-19: 1.3.0: - bug fix in XmmPsf::integratedPsfImage()
//								 (image was not normalized)
//		UL, 2000-11-22: 1.4.0: - added dispersionFigureDistribution()
//		UL, 2000-12-15: 1.5.0: - fixed failing reading of PSF maps when
//								 BETAPSF table is present
//		UL, 2000-12-20: 1.6.0: - dispersionFigureDistribution: removed
//								 constraint: 0<=inChannelOffset<1
//		UL, 2001-01-31: 1.7.0: - changed integration limits in
//								 dispersionFigureDistribution back to
//								 [n-.5*delta, n+.5*delta] (they were
//							     [n, n+delta] in cal>=3.92)
//		UL, 2001-05-11: 1.8.0: - fixed SPR 2027: Normalization of dispersed
//								 PSF distribution wrong
//		UL, 2001-05-15: 1.8.1: - commented out accidentally left over
//								 diagnostic output statements
//		UL, 2001-11-01: 1.9.0: - added implementation of new class
//								 CnrAnalyticPsf
//		UL, 2001-12-21: 1.10.0:- if CCF driver table (KING-PARAM) for
//								 CnrAnalyticPsf (accuracyLevel=EXTENDED) is
//								 not found revert to HIGH accuracy (used to
//								 revert to MEDIUM)
//		UL, 2002-01-08: 1.11.0:- set accuracyLevel to HIGH in state if
//								 EXTENDED PSF model is missing from CCF
//		UL, 2002-01-29: 1.11.1:- bug fix in Interpolated2DParamDataServer
//								 constructor (cdp2 was deleted twice - caused
//								 crash under Linux)
//	   JDR, 2002-06-13: 1.12.0:- dispersionFigureDistribution: changes to
//								 alignment of first moment in returned
//								 distribution
//
//	AUTHOR:
//		U. Lammers (UL), Wed Sep  6 11:14:59 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd, et al.
//		[3]: In Flight Calibration of the PSF for the MOS1 and MOS2 cameras,
//			 S. Ghizzardi, CNR Milano, EPIC-MCT-TN-001, 
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/PsfDataServer.cc,v 1.66 2013/04/03 15:34:37 rsaxton Exp $
//

#include <cmath>
#include <algorithm>
//#include <iomanip> //RDS - remove

#include "Tracer.h"
#include "CalOalUtility.h"
#include "CalServer.h"
#include "PsfDataServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( PSF )----------------------------------------------------------
//
Psf::Psf() : _E(0.), _theta(0.), _phi(0.), _posang(0),
			 _eVal(ValidityInterval(100., 15000.)),
			 _thetaVal(ValidityInterval(0., 900.)),
			 _phiVal(ValidityInterval(0., M_TWO_PI))
{
} // Psf::Psf()


const Psf &Psf::setParameters(double E, double theta, double phi)
{
	_E = E;
	if (E<_eVal.lo() || E>_eVal.hi())
		errstr<<"Request PSF at energy E="<<E<<" eV which is outside the "
				"validity range ["<<_eVal.lo()<<", "<<_eVal.hi()<<"]"
			  <<Sas::warning("EnergyOutsideValidityRange");

	_theta = theta/3600. * DEG_TO_RAD;	//	arcsec -> rad
	if (theta<_thetaVal.lo() || theta>_thetaVal.hi())
		errstr<<"Request PSF at off-angle theta="<<theta<<" arcsec which is "
				"outside the validity range ["<<_thetaVal.lo()<<", "
			  <<_thetaVal.hi()<<"]"
			  <<Sas::warning("OffaxisAngleOutsideValidityRange");

	_phi = fmod(phi, M_TWO_PI);
	if (_phi < 0.)
		_phi += M_TWO_PI;
	if (_phi<_phiVal.lo() || phi>_phiVal.hi())
		errstr<<"Request PSF at azimuth-angle phi="<<phi*RAD_TO_DEG<<" deg "
				"which is outside the validity range ["
			  <<_phiVal.lo()*RAD_TO_DEG<<", "<<_phiVal.hi()*RAD_TO_DEG<<"]"
			  <<Sas::warning("AzimuthAngleOutsideValidityRange");
	return *this;
} // Psf::setParameters()


/* Sets the position angle */
const Psf &Psf::setPosAngle(double pa) 
{

 // Mutator to change the position angle from the default of 0. 
 // This also adds to the rotation of the PSF and so effects PHI.
 // NB: We may have a problem here if the PSF becomes truly dependent
 // on phi !

 // The position angle input is measured anti-clockwise from the north 
 // Experimentally we find that the position angle as used here to 
 // change the orientation of the PSF elements goes in the opposite direction.

  //_posang = M_PI / 2.0 - pa;
  _posang = pa;

  //cout<<"position angle: "<<pa<<endl; // Just to use pa

  return *this;
}

/* AI gcc4 porting
 *  ‘BinSize’ in class ‘Psf’ does not name a type
 */
Psf::ValidBinSize Psf::binSize() const
{
        return _binSize;
} // Psf::binSize()

/* AI gcc4 porting
 *  expected unqualified-id before ‘&’ token, and, expected ‘,’ or ‘...’ before ‘&’ token
 *  and, prototype for ‘const Psf& Psf::binSize() const’ does not match any in class ‘Psf’
 *  => Psf::ValidBinSize
 */
const Psf &Psf::binSize(const Psf::ValidBinSize &binSize) const
{
        if (binSizeChangeable())
                _binSize = binSize;
        else
                errstr<<"Bin size of this PSF implementation is not changeable"
                          <<Sas::error("BinSizeUnchangeable");

        return *this;
} // Psf::binSize()


Psf::ValidityInterval Psf::eValidity() const { return _eVal; }
Psf::ValidityInterval Psf::thetaValidity() const { return _thetaVal; }
Psf::ValidityInterval Psf::phiValidity() const { return _phiVal; }

const Psf &Psf::atOrigin(double x, double y) const
{
	_xOffset = x;
	_yOffset = y;

	return *this;
} // Psf::atOrigin()

template<class T>
static T value(const Block *blk, const string attName, T dftVal)
{
	if (blk->hasAttribute(attName))
		return blk->attribute(attName)->asReal();

	return dftVal;
} // value()

//
//	XmmPsf
//
XmmPsf::XmmPsf(const Table *alignmentParamTable,
			   const Block *binningParameterBlock) : Psf()
{
    MiscDataServer *mds = 0; mds = calServer.getAtom(mds);
    _F = mds->get("FOCAL_LENGTH");
	_px = binningParameterBlock->attribute("Y_PIXSZ")->asReal();
	_py = binningParameterBlock->attribute("Z_PIXSZ")->asReal();
	_nx = binningParameterBlock->attribute("Y_EXTENT")->asReal();
	_ny = binningParameterBlock->attribute("Z_EXTENT")->asReal();

	eValidity(Psf::ValidityInterval(
				value(binningParameterBlock, "E_MIN", 100.),
				value(binningParameterBlock, "E_MAX", 15000.)));
	thetaValidity(Psf::ValidityInterval(
				value(binningParameterBlock, "THET_MIN", 0.) * 60.,
				value(binningParameterBlock, "THET_MAX", 15.) * 60.));
	phiValidity(Psf::ValidityInterval(
				value(binningParameterBlock, "PHI_MIN", 0.),
				value(binningParameterBlock, "PHI_MAX", M_TWO_PI)));

	//
	//	init PSF sampling bin size to intrinsic one
	//
        /* AI gcc4 porting
         * error: missing template arguments before '(' token => ValidBinSize
         */
        binSize(ValidBinSize(_px, _py));

	ColumnData
		*cdTheta = alignmentParamTable->column("THETA")->data(),
		*cdDecenterY = alignmentParamTable->column("DECENTER_Y")->data(),
		*cdDecenterZ = alignmentParamTable->column("DECENTER_Z")->data(),
		*cdDefocus = alignmentParamTable->column("DEFOCUS")->data();
	real32 *theta = cdTheta->real32Data(),
		   *dY = cdDecenterY->real32Data(),
		   *dZ = cdDecenterZ->real32Data(),
		   *dF = cdDefocus->real32Data();

	unsigned N = alignmentParamTable->rows();
	vector<double> thetav(N), dYv(N), dZv(N), dFv(N);
	for (unsigned n=0; n<N; ++n) {
		thetav[n] = theta[n];
		dYv[n] = dY[n];
		dZv[n] = dZ[n];
		dFv[n] = dF[n];
	}
	delete cdTheta;
	delete cdDecenterY;
	delete cdDecenterZ;
	delete cdDefocus;

	_deltaYInter = InterpolatedCurve(thetav, dYv);
	_deltaZInter = InterpolatedCurve(thetav, dZv);
	_deltaFInter = InterpolatedCurve(thetav, dFv);
} // XmmPsf::XmmPsf()


const Psf &XmmPsf::setParameters(double E, double theta, double phi)
{
	Psf::setParameters(E, theta, phi);

	_oneOverPx	= 1./_px;
	_oneOverPy	= 1./_py;
	_sinPhi		= sin(this->phi());
	_cosPhi		= cos(this->phi());
	//_sinPhi		= sin(this->phi() * -1.0);
	//_cosPhi		= cos(this->phi() * -1.0);
	_Fbar		= (_F + _deltaFInter(this->theta())) * tan(this->theta());
	_deltaX		= _deltaYInter(this->theta());
	_deltaY		= _deltaZInter(this->theta());

	FocalPlanePosition fp = barycenter();
	_baryX = fp.first;
	_baryY = fp.second;
	atOrigin(0., 0.);

	return *this;
} // XmmPsf::setParameters()

/* Sets the position angle - input in Radians */
const Psf &XmmPsf::setPosAngle(double pa) 
{

 // Mutator to change the position angle from the default of 0. 
 // This also adds to the rotation of the PSF and so effects PHI.
 // NB: We may have a problem here if the PSF becomes truly dependent
 // on phi !
  Psf::setPosAngle(pa);

  //double newphi = this->phi() + this->posang() + M_PI / 2.0;
 // Position Angle rotates anti-clockwise hence we subtract it from phi
 // which is rotating clockwise in the image coordinates that we
 // are using here.
  double newphi = this->phi() - this->posang();

  //cout<<"NEWPHI: "<<this->phi()*180.0/3.1415<<" "<<newphi<<endl;

  if (newphi > (M_PI * 2.0) ) { newphi = newphi - M_PI * 2.0;}
  if (newphi < (M_PI * -2.0) ) { newphi = newphi + M_PI * 2.0;}

  //cout<<"NEWPHI2: "<<newphi<<endl;

 // Need to convert theta back to arcsecs
  double theta = this->theta()*3600. / DEG_TO_RAD;
  setParameters(this->E(), theta, newphi);

  //cout<<"position angle: "<<pa<<endl; // Just to use pa

  return *this;
} // Psf::setPosAngle()




inline pair<Image<float>, double> XmmPsf::integratedPsfImage(const PsfWindow &window) const
{
	double binX = binSize().binX(),
		   binY = binSize().binY();
	int Nx = UNT((window.halfX()-.5*binX)/binX + .5),
		Ny = UNT((window.halfY()-.5*binY)/binY + .5);
	Image<float> map(2*Nx+1, 2*Ny+1);

	double tot = 0.;
	for (int nx=-Nx; nx<=Nx; ++nx)
		for (int ny=-Ny; ny<=Ny; ++ny) {
			double x = DBL(nx)*binX + _baryX + _xOffset,
				   y = DBL(ny)*binY + _baryY + _yOffset;
                      /* Set the pixel value to the value of the PSF
			 value in the CCF rotated and translated to the
                         required phi */
			map[nx+Nx][ny+Ny] = (*this)(x, y);
			tot += map[nx+Nx][ny+Ny];
		}
        //cout<<"From CAL[60,59]: "<<map[60][59]<<" "<<map[598][818]<<endl;

       /* Add in the spoke pattern if required. We have to do this
	  explicitely here because the (this)(x,y) construction is
          already rotated and we do not want the spokes rotated */
        Image<float> omap = this->addSpokes(map);
        //Image<float> omap = map;   // REMOVE !!!!!!!!!!!!!!!!!!!

	tot *= binX*binY;

	SAS_DEBUG(cout<<"binX/binY: "<<binX<<'/'<<binY<<endl;)
	SAS_DEBUG(cout<<"Total integrated: "<<tot<<' '<<tot<<endl;)

	return PsfPair(omap, tot);
} // XmmPsf::integratedPsfImage()


XmmPsf::PsfWindow XmmPsf::enclosingBox(double fraction,
									   double aspectRatioYoverX) const
{
	if (fraction >= 1.)
		errstr<<"Cannot determine PSF bounding box that includes "
			  <<100.*fraction<<"% of the energy"
			  <<Sas::error("InvalidInput");

	_binX = binSize().binX();
	_binY = binSize().binY();
	_binArea = _binX * _binY;

	int	maxX = 0,
		maxY = 0;
	double totPsf = intPsf(0, 0);
	while (totPsf < fraction) {
		//
		//	expand box in X by one pixel
		//
		++maxX;
		for (int iy=-maxY; iy<=maxY; ++iy)
			totPsf += intPsf(maxX, iy) + intPsf(-maxX, iy);

		//
		//	expand box in Y by aspectRatioYoverX pixels
		//
		int newMaxY = INT(.5*(DBL(2*maxX+1) * aspectRatioYoverX - 1.) + .5);
		for (int iy=maxY+1; iy<=newMaxY; ++iy)
			for (int ix=maxX; ix>=-maxX; --ix)
				totPsf += intPsf(ix, iy) + intPsf(ix, -iy);
		maxY = newMaxY;
	}

	return PsfWindow(maxX * _binX, maxY * _binY);
} // XmmPsf::enclosingBox()


double XmmPsf::enclosingCircle(double fraction) const
{
	//	in 0'th approximation just return radius of circle with same
	//	area as corresponding square box

	return 2.*enclosingBox(fraction).halfX()/sqrt(M_PI);
} // XmmPsf::enclosingEnergy()


double XmmPsf::enboxedEnergy(const PsfWindow &window) const
{
	return integratedPsfImage(window).second;
} // XmmPsf::enboxedEnergy()


double XmmPsf::encircledEnergy(double radius) const
{
	//	approximation: integrate PSF in square with area equal to enclosing
	//	circle
	//double aHalf = 5*radius*sqrt(M_PI);
	double aHalf = 0.5*radius*sqrt(M_PI);

	return min(enboxedEnergy(PsfWindow(aHalf, aHalf)), 1.);
} // XmmPsf::encircledEnergy()


inline double XmmPsf::intPsf(int ix, int iy) const
{
	return _binArea * (*this)(DBL(ix)*_binX + _baryX + _xOffset,
							  DBL(iy)*_binY + _baryY + _yOffset);
} // XmmPsf::intPsf()


double XmmPsf::xCc2(double xPsf, double yPsf) const
{
	double xb = xPsf - .5*_px*(_nx-1.),
		   yb = yPsf - .5*_py*(_ny-1.);

	return (xb + _Fbar) * _cosPhi + yb * _sinPhi + _deltaX;
} // XmmPsf::xCc2()


double XmmPsf::yCc2(double xPsf, double yPsf) const
{
	double xb = _px*(xPsf - .5*(_nx-1.)),
		   yb = _py*(yPsf - .5*(_ny-1.));

	return (-xb + _Fbar) * _sinPhi + yb * _cosPhi + _deltaY;
} // XmmPsf::yCc2()


FocalPlanePosition XmmPsf::barycenter() const
{
#if 0	
	float theta = this->theta() * RAD_TO_DEG * 3600.,
		  phi = this->phi();

	GeometryDataServer *cdf = 0; cdf = calServer.getAtom(cdf);
	CalReal32Vector thetav(&theta, 1),
					phiv(&phi, 1);

	float baryX, baryY, baryZ;
	XmmCartesianPointList pl =
		cdf->telCoordList(thetav, phiv).toCamCoord2(&baryX, &baryY, &baryZ);
#endif

      // CHANGED THESE TWO LINES !!
	//double phi = this->phi();
	//return FocalPlanePosition(_Fbar*cos(phi), _Fbar*sin(phi));
	return FocalPlanePosition(_Fbar*_cosPhi, _Fbar*_sinPhi);
} // XmmPsf::barycenter()

/* Method xPsf: Calculate the X coordinate of the PSF image, translated to
the barycentre and rotated clockwise with the azimuthal angle */
double XmmPsf::xPsf(double xCc2, double yCc2) const
{
	double x = xCc2 - _deltaX,
		   y = yCc2 - _deltaY,

   // Rotate the point clockwise with Phi
		  xb = x*_cosPhi - y*_sinPhi +_Fbar*(_sinPhi*_sinPhi-_cosPhi*_cosPhi);
	//	   xb = x*_cosPhi + y*_sinPhi +_Fbar*(_cosPhi*_cosPhi-_sinPhi*_sinPhi);


	return (xb + .5*_px*(_nx-1.))/_px;
} // XmmPsf::xPsf()


/* Method xPsf: Calculate the Y coordinate of the PSF image, translated to
the barycentre and rotated clockwise with the azimuthal angle */
double XmmPsf::yPsf(double xCc2, double yCc2) const
{
	double x = xCc2 - _deltaX,
		   y = yCc2 - _deltaY,

   // Rotate the point clockwise with Phi
	yb = x*_sinPhi + y*_cosPhi - 2.*_Fbar*_sinPhi*_cosPhi;
	//	   yb = -x*_sinPhi + y*_cosPhi + 2.*_Fbar*_sinPhi*_cosPhi;

	return (yb + .5*_py*(_ny-1.))/_py;
} // XmmPsf::yPsf()


//
//	Interpolated2DParamDataServer
//
Interpolated2DParamDataServer::Interpolated2DParamDataServer(
												const Table *fitParamTable,
												const string &p1ColName,
												const string &p2ColName,
												const string &paramColName)
{
	ColumnData *cdp1 = fitParamTable->column(p1ColName)->data(),
			   *cdp2 = fitParamTable->column(p2ColName)->data(),
			   *cdParam = fitParamTable->column(paramColName)->data();

	if (cdp1->elements() != 1 ||
		cdp2->elements() != 1)
		errstr<<cdp1->parent()->qualifiedName()<<'/'
			  <<cdp2->parent()->qualifiedName()
			  <<" need to be scalar real32 columns"
			  <<Sas::error("UnexpectedCCFstructure");

	real32 *p1 = cdp1->real32Data(),
		   *p2 = cdp2->real32Data(),
		   *param = cdParam->real32Data();

	unsigned N = fitParamTable->rows();
	vector<float> p1v(p1, p1+N), p2v(p2, p2+N);
	vector<float>::iterator vi = unique(p1v.begin(), p1v.end());
	p1v.erase(vi, p1v.end());

	sort(p2v.begin(), p2v.end());
	vi = unique(p2v.begin(), p2v.end());
	p2v.erase(vi, p2v.end());

	vector<float> p(N);

        //cout<<"Rows: "<<p1v.size()<<" "<<p2v.size()<<" "<<p.size()<<endl;

	assert(p1v.size() * p2v.size() == p.size());
	_nParams = cdParam->elements();
	for (unsigned i=0; i<_nParams; ++i) {
		for (unsigned j=0; j<N; ++j)
			p[j] = param[cdParam->position(j, i)];

		_interpolatedSurfaces.push_back(new InterpolatedSurface(p2v, p1v, p));
		//	yes, p2 then p1; row-wise order of p

		SAS_DEBUG(
			cout<<"i="<<i<<endl;
			for (unsigned ii=0; ii<p1v.size(); ++ii) {
				for (unsigned jj=0; jj<p2v.size(); ++jj)
					cout<<(*_interpolatedSurfaces[i])(p2v[jj], p1v[ii])<<' ';
				cout<<endl;
			});
	}

	delete cdp1;
	delete cdp2;
	delete cdParam;
} // Interpolated2DParamDataServer::Interpolated2DParamDataServer()


Interpolated2DParamDataServer::~Interpolated2DParamDataServer()
{
	for (unsigned n=0; n<_interpolatedSurfaces.size();
		 delete _interpolatedSurfaces[n++]){};
} // Interpolated2DParamDataServer::~Interpolated2DParamDataServer()

//
//	LinInterpolated2DParamDataServer
//
LinInterpolated2DParamDataServer::LinInterpolated2DParamDataServer(
												const Table *fitParamTable,
												const string &p1ColName,
												const string &p2ColName,
												const string &paramColName)
{
	ColumnData *cdp1 = fitParamTable->column(p1ColName)->data(),
			   *cdp2 = fitParamTable->column(p2ColName)->data(),
			   *cdParam = fitParamTable->column(paramColName)->data();

	if (cdp1->elements() != 1 ||
		cdp2->elements() != 1)
		errstr<<cdp1->parent()->qualifiedName()<<'/'
			  <<cdp2->parent()->qualifiedName()
			  <<" need to be scalar real32 columns"
			  <<Sas::error("UnexpectedCCFstructure");

	real32 *p1 = cdp1->real32Data(),
		   *p2 = cdp2->real32Data(),
		   *param = cdParam->real32Data();

	unsigned N = fitParamTable->rows();
	vector<float> p1v(p1, p1+N), p2v(p2, p2+N);
	vector<float>::iterator vi = unique(p1v.begin(), p1v.end());
	p1v.erase(vi, p1v.end());

	sort(p2v.begin(), p2v.end());
	vi = unique(p2v.begin(), p2v.end());
	p2v.erase(vi, p2v.end());

	vector<float> p(N);

        //cout<<"Rows: "<<p1v.size()<<" "<<p2v.size()<<" "<<p.size()<<endl;

	assert(p1v.size() * p2v.size() == p.size());
	_nParams = cdParam->elements();
	for (unsigned i=0; i<_nParams; ++i) {
		for (unsigned j=0; j<N; ++j)
			p[j] = param[cdParam->position(j, i)];

		_interpolatedSurfaces.push_back(new LinInterpolatedSurface(p2v, p1v, p));
		//	yes, p2 then p1; row-wise order of p

		SAS_DEBUG(
			cout<<"i="<<i<<endl;
			for (unsigned ii=0; ii<p1v.size(); ++ii) {
				for (unsigned jj=0; jj<p2v.size(); ++jj)
					cout<<(*_interpolatedSurfaces[i])(p2v[jj], p1v[ii])<<' ';
				cout<<endl;
			});
	}

	delete cdp1;
	delete cdp2;
	delete cdParam;
} // Interpolated2DParamDataServer::Interpolated2DParamDataServer()


LinInterpolated2DParamDataServer::~LinInterpolated2DParamDataServer()
{
	for (unsigned n=0; n<_interpolatedSurfaces.size();
		 delete _interpolatedSurfaces[n++]){};
} // Interpolated2DParamDataServer::~Interpolated2DParamDataServer()

//
//	CpInterpolated2DParamDataServer
//
CpInterpolated2DParamDataServer::CpInterpolated2DParamDataServer(
												const Table *fitParamTable,
												const string &p1ColName,
												const string &p2ColName,
												const string &paramColName)
{
	ColumnData *cdp1 = fitParamTable->column(p1ColName)->data(),
			   *cdp2 = fitParamTable->column(p2ColName)->data(),
			   *cdParam = fitParamTable->column(paramColName)->data();

	if (cdp1->elements() != 1 ||
		cdp2->elements() != 1)
		errstr<<cdp1->parent()->qualifiedName()<<'/'
			  <<cdp2->parent()->qualifiedName()
			  <<" need to be scalar real32 columns"
			  <<Sas::error("UnexpectedCCFstructure");

	real32 *p1 = cdp1->real32Data(),
		   *p2 = cdp2->real32Data(),
		   *param = cdParam->real32Data();

	unsigned N = fitParamTable->rows();
	vector<float> p1v(p1, p1+N), p2v(p2, p2+N);
	vector<float>::iterator vi = unique(p1v.begin(), p1v.end());
	p1v.erase(vi, p1v.end());

	sort(p2v.begin(), p2v.end());
	vi = unique(p2v.begin(), p2v.end());
	p2v.erase(vi, p2v.end());

	vector<float> p(N);

        //cout<<"Rows: "<<p1v.size()<<" "<<p2v.size()<<" "<<p.size()<<endl;

	assert(p1v.size() * p2v.size() == p.size());
	_nParams = cdParam->elements();
	for (unsigned i=0; i<_nParams; ++i) {
		for (unsigned j=0; j<N; ++j)
			p[j] = param[cdParam->position(j, i)];

		_interpolatedSurfaces.push_back(new CpInterpolatedSurface(p2v, p1v, p));
		//	yes, p2 then p1; row-wise order of p

		SAS_DEBUG(
			cout<<"i="<<i<<endl;
			for (unsigned ii=0; ii<p1v.size(); ++ii) {
				for (unsigned jj=0; jj<p2v.size(); ++jj)
					cout<<(*_interpolatedSurfaces[i])(p2v[jj], p1v[ii])<<' ';
				cout<<endl;
			});
	}

	delete cdp1;
	delete cdp2;
	delete cdParam;
} // Interpolated2DParamDataServer::Interpolated2DParamDataServer()


CpInterpolated2DParamDataServer::~CpInterpolated2DParamDataServer()
{
	for (unsigned n=0; n<_interpolatedSurfaces.size();
		 delete _interpolatedSurfaces[n++]){};
} // Interpolated2DParamDataServer::~Interpolated2DParamDataServer()


//
//	CslAnalyticPsfLow
//
CslAnalyticPsfLow::CslAnalyticPsfLow(const Table *alignmentParamTable,
									 const Table *fitParamTable)
	: XmmPsf(alignmentParamTable, fitParamTable),
	  _paramDataServer(fitParamTable, "ENERGY", "THETA", "PARAMS")
{
	_param.resize(_paramDataServer.parameters());
	_norm = 1./px()/py();
} // CslAnalyticPsfLow::CslAnalyticPsfLow()


const Psf &CslAnalyticPsfLow::setParameters(double E, double theta, double phi)
{
	XmmPsf::setParameters(E, theta, phi);
	double e = this->E(),
		   t = this->theta();

	_param = _paramDataServer(t, e);
	if (_param.size() != 6)
		errstr<<"Expected 6 parameters to drive CslAnalyticPsfLow but there "
				"are "<<_param.size()<<" - this is a CCF problem."
			  <<Sas::error("WrongNumberOfParameters");

	_norm = 1./(M_TWO_PI*_param[2]*_param[3]*_param[1]*px()*py());

	return *this;
} // CslAnalyticPsfLow::setParameters()


double CslAnalyticPsfLow::pointValue(double x, double y) const
{
	return _norm * _param[1] *
				exp(-.5*(CalOalUtility::SQR((x-_param[4])/_param[2]) +
						 CalOalUtility::SQR((y-_param[5])/_param[3])));
} // CslAnalyticPsfLow::pointValue()


CslAnalyticPsfLow::~CslAnalyticPsfLow()
{
} // CslAnalyticPsfLow::~CslAnalyticPsfLow()


//
//	CslAnalyticPsfMedium
//
CslAnalyticPsfMedium::CslAnalyticPsfMedium(const Table *alignmentParamTable,
										   const Table *fitParamTable)
	: XmmPsf(alignmentParamTable, fitParamTable)
{
	ColumnData *cdParam = fitParamTable->column("PARAMS")->data();
	real32 *param = cdParam->real32Data();
	unsigned Np = cdParam->elements();
	_Ap.resize(Np);
	_Bp.resize(Np);
	_Cp.resize(Np);

	assert(fitParamTable->rows()==3);
	for (unsigned n=0; n<Np; ++n) {
		_Ap[n] = param[cdParam->position(0, n)];
		_Bp[n] = param[cdParam->position(1, n)];
		_Cp[n] = param[cdParam->position(2, n)];
	}

	delete cdParam;

	_norm = 1./fitParamTable->attribute("NORM")->asReal()/px()/py();
} // CslAnalyticPsfMedium::CslAnalyticPsfMedium()


const Psf &CslAnalyticPsfMedium::setParameters(double E,
											   double theta, double phi)
{
	XmmPsf::setParameters(E, theta, phi);
	return *this;
} // CslAnalytic::setParameters()


double CslAnalyticPsfMedium::pointValue(double x, double y) const
{
	double phi = atan2(x-_Ap[1], y-_Ap[2]);
	return (_Ap[0] * exp(-(CalOalUtility::SQR(x-_Ap[1])/_Ap[3] +
						 CalOalUtility::SQR(y-_Ap[2])/_Ap[4])) *
					(1.+_Ap[5] * cos(3.*phi+_Ap[6])) +
		   _Bp[0] * exp(-(CalOalUtility::SQR(x-_Bp[1])/_Bp[3] +
						 CalOalUtility::SQR(y-_Bp[2])/_Bp[4])) *
					(1.+_Bp[5] * cos(3.*phi+_Bp[6])) +
		   _Cp[0] * exp(-(CalOalUtility::SQR(x-_Cp[1])/_Cp[3] +
						 CalOalUtility::SQR(y-_Cp[2])/_Cp[4])) *
					(1.+_Cp[5] * cos(16.*phi+_Cp[6])))*_norm;
} // CslAnalyticPsfMedium::pointValue()


CslAnalyticPsfMedium::~CslAnalyticPsfMedium()
{
} // CslAnalyticMedium::~CslAnalyticPsfMedium()


//
//	CnrAnalyticPsf - see [3] for details
//
CnrAnalyticPsf::CnrAnalyticPsf(const Table *alignmentParamTable,
							   const Table *fitParamTable)
	: XmmPsf(alignmentParamTable, fitParamTable),
	  _paramDataServer(fitParamTable, "ENERGY", "THETA", "PARAMS")
{
	//	read normalization radius parameter [arcmin]
	const Attribute *att = fitParamTable->attribute("NORM_AT");
	assert(att->units() == "arcmin");
	_normAt = att->asReal() * 60.;	//	arcmin->arcsec
} // CnrAnalyticPsf::CnrAnalyticPsf()


const Psf &CnrAnalyticPsf::setParameters(double E,
											   double theta, double phi)
{
	XmmPsf::setParameters(E, theta, phi);
	double e = this->E(),
		   t = this->theta();

	_param = _paramDataServer(t, e);
	if (_param.size() != 2)
		errstr<<"Expected 2 parameters to drive CnrAnalyticPsf but there "
				"are "<<_param.size()<<" - this is a CCF problem."
			  <<Sas::error("WrongNumberOfParameters");

	_rc2 = CalOalUtility::SQR(_param[0]);
	_alpha1 = _param[1]-1.;
	_fac = M_PI*_rc2/_alpha1;
	_norm = 1./(_fac*(1.-1./pow(1.+CalOalUtility::SQR(_normAt)/_rc2,
								_alpha1)));

        //cout<<"Params: "<<_param[0]<<" "<<_param[1]<<" "<<_param[2]<<endl;
        //cout<<"NORM: "<<_norm<<endl;


	return *this;
} // CslAnalytic::setParameters()


double CnrAnalyticPsf::pointValue(double x, double y) const
{
	//	see [3], Eq. (2)
        //cout<<"ERE: "<<x<<" "<<y<<" "<<_rc2<<" "<<_norm/pow(1.+(CalOalUtility::SQR(x)+CalOalUtility::SQR(y))/_rc2,_param[1])<<endl;
	return _norm/pow(1.+(CalOalUtility::SQR(x)+CalOalUtility::SQR(y))/_rc2,
					 _param[1]);
} // CnrAnalyticPsf::pointValue()


double CnrAnalyticPsf::encircledEnergy(double radius) const
{
	if (radius < 0.) {
		errstr<<"Given radius for encircled energy computation is negative"
			  <<Sas::warning("NegativeRadius");
		return 0.;
	}

	//	see [3], Eq. (10)
	radius /= px();	//	mm->arcsec
	return min(_norm*_fac*(1.-1./pow(1.+CalOalUtility::SQR(radius)/_rc2,
			   _alpha1)), 1.);
} // CnrAnalyticPsf::encircledEnergy()


double CnrAnalyticPsf::enclosingCircle(double fraction) const
{
	if (fraction < 0. || fraction >= 1) {
		errstr<<"Given encircled energy fraction "<<fraction<<" is outside "
				"allowed interval [0, 1]"
			  <<Sas::warning("IllegalInputValue");
		return 0.;
	}

	//	see [3], Eq. (11)
	return _param[0]*sqrt(pow(1.-fraction/_fac/_norm, -1./_alpha1) - 1.)*px();
} // CnrAnalyticPsf::enclosingCircle()


CnrAnalyticPsf::~CnrAnalyticPsf()
{
} // CnrAnalyticPsf::~CnrAnalyticPsf()


//
//	EbetaAnalyticPsf - see [?] for details
//
EbetaAnalyticPsf::EbetaAnalyticPsf(const Table *alignmentParamTable,
							   const Table *fitParamTable)
	: XmmPsf(alignmentParamTable, fitParamTable),
	  _paramDataServer(fitParamTable, "ENERGY", "THETA", "PARAMS")
	  // _paramDataServer(fitParamTable, "ENERGY", "THETA", "PHI", "PARAMS")
          // Use PHI if necessary.
{
	//	read normalization radius parameter [arcmin]
	const Attribute *att = fitParamTable->attribute("NORM_AT");
	assert(att->units() == "arcmin");
	_normAt = att->asReal() * 60.;	//	arcmin->arcsec

        // This version needs to have the RD1,2,3 values to calculate the
        // spokes - issue an error if not found
        if (! fitParamTable->hasAttribute("RD1"))
           errstr<<"Keyword RD1 not found in ELLBETA PSF extension, "<<
                   "please update your CCF"<<
                   Sas::error("UnexpectedCCFstructure");

        //      read spoke specific parameters
        _ophi = fitParamTable->attribute("OPHI")->asReal();
        _nsp = fitParamTable->attribute("NSP")->asInt();
        _rthresh = fitParamTable->attribute("RTHRESH")->asReal();
        _changefrac = fitParamTable->attribute("CHNGFRAC")->asReal();
        _m1frac= fitParamTable->attribute("M1FRAC")->asReal();
        _m2frac= fitParamTable->attribute("M2FRAC")->asReal();
        _secfrac= fitParamTable->attribute("SECFRAC")->asReal();
        _u = fitParamTable->attribute("U")->asReal();
        _v = fitParamTable->attribute("V")->asReal();
//        _rlo = fitParamTable->attribute("RLO")->asReal();
//        _rhi = fitParamTable->attribute("RHI")->asReal();
        _rd1 = fitParamTable->attribute("RD1")->asReal();
        _rd2 = fitParamTable->attribute("RD2")->asReal();
        _rd3 = fitParamTable->attribute("RD3")->asReal();

        // Attempt to read coefficients which describe a radial offset
        // of the PSF centre. Otherwise set to zero.
        if (fitParamTable->hasAttribute("OFF_0")) {
          _cenOffCoeff.push_back(fitParamTable->attribute("OFF_0")->asReal()); 
          _cenOffCoeff.push_back(fitParamTable->attribute("OFF_1")->asReal()); 
          _cenOffCoeff.push_back(fitParamTable->attribute("OFF_2")->asReal()); 
          _cenOffCoeff.push_back(fitParamTable->attribute("OFF_3")->asReal()); 
        } else {
           for (unsigned i=0; i<=3; ++i) 
                 _cenOffCoeff.push_back(0.0);
        }
           
} // EbetaAnalyticPsf::EbetaAnalyticPsf()


const Psf &EbetaAnalyticPsf::setParameters(double E,
											   double theta, double phi)
{
	XmmPsf::setParameters(E, theta, phi);
	double e = this->E(),
		   t = this->theta();
                   //p = this->phi();

	_param = _paramDataServer(t, e);
	if (_param.size() != 3 && _param.size() != 5)
		errstr<<"Expected 5 parameters to drive EbetaAnalyticPsf but there "
				"are "<<_param.size()<<" - this is a CCF problem."
			  <<Sas::error("WrongNumberOfParameters");

        //cout<<"Params: "<<theta<<" "<<_param[0]<<" "<<_param[1]<<" "<<_param[2]<<
          //  " "<<_param[3]<<" "<<_param[4]<<" E: "<<e<<" "<<t<<endl;

       /* We cater for the case where the full King and Gaussian parameters
          are in the file and the earlier case where only the elliptical
          King function params were included. */

       // Set the elliptical King function params 
	_rc2 = CalOalUtility::SQR(_param[0]);
	_alpha1 = _param[1]-1.;
        _ellipity = _param[2];

       // Set the normalisation for the KING fn.
       // Differs from circular King function by (1-ellip) to account for 
       // the area of the elliptical annulus
	//_fac = M_PI*_rc2 /_alpha1;
	_fac = M_PI*_rc2 * (1.0-_ellipity)/_alpha1;
	_norm = 1./(_fac*(1.-1./pow(1.+CalOalUtility::SQR(_normAt)/_rc2,
							 	_alpha1)));

       // Set the Gaussian function params 
        if (_param.size()  == 5) {
         // Set the FWHM of the Gaussian component in arcseconds 
 	   _gfwhm = _param[3];
         // Set the relative normalisation of the Gaussian. Don't let
         // it be interpolated to a -ve value.
 	   _gnorm = max(_param[4],0.0);  
        } else {
           _gfwhm = -999.0;
           _gnorm = 0.0;
        }

       // REMOVE !!!!!!!!!!!!!!
       // _gnorm = 0.0;

       // Set the normalisation of the area under the Gaussian so that
       // It integrates to 1.0
        double ln2 = 0.693147;
        _fourln2 = 4.0 * ln2; 

       // This is the sqrt(2*pi*sigma^2) gaussian integral
        double gauss_factor = 
                    sqrt( ( M_PI * _gfwhm * _gfwhm ) / (4.0 * _fourln2) );

       // This is something which makes the relative normalisation 
       // come out about right compared with integrating the images 
       // - dont know why
        double fudge = sqrt(M_PI * _gfwhm * _gfwhm / ln2);
    //    double fudge=1.0; // REMOVE !!!!!

        _gauss_area_norm = _norm * _gnorm * gauss_factor * fudge;

        _gratio = _norm * _gnorm;

        //cout<<"NORMS: "<<_norm<<" "<<_gnorm<<" "<<pow( ( M_PI * _gfwhm * _gfwhm ) / (4.0 * _fourln2), 0.5)<<" "<<_gauss_area_norm<<endl;

       // Renormalise so that the King + Gaussian integrate to 1.0
        double totnorm = 1.0 + _gauss_area_norm;
        //double totnorm = 1.0;  // SORT THIS OUT !!!!!!!!!!
        _norm /= totnorm;
        _gnorm /= totnorm;
        _gratio /= totnorm;
        _gauss_area_norm /= totnorm;

       // Calculate some useful local variables
        _elipdiv = (1-_ellipity)*(1-_ellipity);

       // Avoid /0 error
        if (_gfwhm > 0) {
          _gausexp = -1.0 * _fourln2 / (_gfwhm*_gfwhm);
        } else {
          _gausexp = 0.0;
        }

        //cout<<"FWHM: "<<_gfwhm<<" "<<_gausexp<<endl;
        //cout<<"NORM: "<<_norm<<" "<<fudge<<" "<<_gratio<<" "<<totnorm<<" "<<_gnorm<<endl;


	return *this;
} // EbetaAnalytic::setParameters()


double EbetaAnalyticPsf::pointValue(double x, double y) const
{
	//	King profile + Gaussian 
        // Returns the point value of the function integrated over
        // a 1"x1" area.

        //double r2 = x*x + y*y / _elipdiv; 
        //double r2 = x*x + (y*y) / ( (1-_ellipity)*(1-_ellipity) ); 
        double r2 = x*x + y*y / _elipdiv; 

        /* Return the value of the King function (normalised to 1.0 over
	   the whole image plus a Gaussian (normalised to the King 
	   function at the peak by the factor _gnorm). Use derived
	   variables to speed up the calculation */ 
	 return ( _norm/pow( (1.+r2/_rc2), _param[1] )
                   + _gratio * exp(_gausexp * r2) );   

//	return min( _norm*_fac*(1.-1./pow(1.+CalOalUtility::SQR(radius)/_rc2,
//		    _alpha1)) + _gauss_area_norm * gauss_int, 1.);

       // Extra normalisation by (1-ellipticity) to make image at infinity
       // = 1.0
	 //return ( _norm/pow( (1.+r2/_rc2), _param[1] )
           //        + _gratio * exp(_gausexp * r2) ) / (1.0 - _ellipity);   


	 //return _gratio * exp(_gausexp * r2);
                  //+ _gratio * exp(_gausexp * r2); 
              //   _norm * _gnorm * exp(-1.0 * _fourln2 * r2 / (_gfwhm*_gfwhm) ); 
 
} // EbetaAnalyticPsf::pointValue()


double EbetaAnalyticPsf::encircledEnergy(double radius) const
{
	if (radius < 0.) {
		errstr<<"Given radius for encircled energy computation is negative"
			  <<Sas::warning("NegativeRadius");
		return 0.;
	}

	//	see [3], Eq. (10)
	radius /= px();	//	mm->arcsec

        /* Calculate the integrated King function out to this Radius +
        // The integrated Gaussian out to this radius
        // NB: Under development !!!
        // Calculate the complementary error function using a "chebychev
        // polynomial fit to an inspired guess" - Numerical Recipes p176
        // The t in exp(-t^2) is [ sqrt(_fourln2) * radius / _gfwhm ]
        */
      //double z = radius;
       /* REINSTATE
        double z = sqrt(_fourln2) * radius * px() / _gfwhm;
        double t = 1.0 / (1.0+0.5*z);
        double erfcc=t*exp(-z*z-1.26551223+t*(1.00002368+t*(.37409196+t*
               (.09678418+t*(-.18628806+t*(.27886807+t*(-1.13520398+t*
               (1.48851587+t*(-.82215223+t*.17087277)))))))));
        double gauss_int = 1.0 - erfcc;
        gauss_int = min(1.0, gauss_int);
        gauss_int = max(0.0, gauss_int);
       */ // REINSTATE

        //cout<<"GAUSS_INT: "<<z<<" "<<gauss_int<<" "<<_gauss_area_norm<<" "<<
              //_norm*_fac*(1.-1./pow(1.+CalOalUtility::SQR(radius)/_rc2,
                //    _alpha1))<<" "<<_gauss_area_norm * gauss_int<<endl;

       // This isn't working yet so replace with an integration of the
       // image, which also helps to compare the radial profiles

	// return min( _norm*_fac*(1.-1./pow(1.+CalOalUtility::SQR(radius)/_rc2,
	////    _alpha1)) + _gauss_area_norm * gauss_int, 1.);
	//	    _alpha1)), 1.);

       /* Map the image */
       /*
	double binX = radius;
		   binY = radius;
	int Nx = UNT((window.halfX()-.5*binX)/binX + .5),
		Ny = UNT((window.halfY()-.5*binY)/binY + .5);
	Image<float> map(2*Nx+1, 2*Ny+1);
       */

	double tot = 0.;
        double r2 = radius*radius;
        int subsize=9, sub2=floor(subsize/2);
        //double norm = 0.;

       /* Loop over the eBeta image in 1" bins */
	for (int nx=-INT(radius); nx<=INT(radius); ++nx) {
	    for (int ny=-INT(radius); ny<=INT(radius); ++ny) {

               // Position in PSF image in mm
		double x = DBL(nx) * px() + _baryX + _xOffset,
		       y = DBL(ny) * py() + _baryY + _yOffset;

               // Calculate the x,y value of the closest corner of 
               // this pixel to the PSF centre
                double minx = abs(nx)-0.5, maxx=minx+1.0;
                double miny = abs(ny)-0.5, maxy=miny+1.0;

               // If the pixel falls wholly within the extraction radius
                if ( (maxx*maxx+maxy*maxy) < r2 )
                {
                   tot += (*this)(x, y);
                }
               // If the pixel falls partially within the radius
                else if ( (minx*minx+miny*miny) < r2 )
                {
                  // Divide the pixel into 9*9 subpixels and average
                  // the values within those subpixels
                  double subtot=0.0;

                  for (int ix=-sub2; ix<=sub2; ++ix)
                    for (int iy=-sub2; iy<=sub2; ++iy) 
                     {
                       double newx=DBL(nx)+DBL(ix)/DBL(subsize),
                              newy=DBL(ny)+DBL(iy)/DBL(subsize);

                       if ((newx*newx + newy*newy) < r2)
                       {
                          x = newx * px() + _baryX + _xOffset;
                          y = newy * py() + _baryY + _yOffset;

                          subtot += (*this)(x, y);
//                  cout<<"PIX:"<<nx<<" "<<ny<<" "<<ix<<" "<<iy<<" "<<newx<<" "<<newy<<" "<<subtot<<" "<<r2<<endl;
                       }
                     }

                  tot += subtot / (subsize*subsize);
                }
	    }
	}

        //cout<<"CAL: encirc energy = "<<tot<<" "<<radius<<endl;

        //return (tot/norm);
        return (tot);
      
} // EbetaAnalyticPsf::encircledEnergy()


vector<double> EbetaAnalyticPsf::enclosingCircle(vector<double> fraction) const
{

   /* Calculates the encircled energy in radii of 1 arcsecond
      upto the full integration radius (normat). Then returns the
      enclosing radii for a set of encircled energy fractions in mm*/

   /* Set the max radius as the defined integration radius */
       int maxrad = _normAt;  // in arcsec
       //int maxrad = 300;  // in arcsec - 

   /* Create a vector to hold the integrated counts in a circle
      of radius r */ 

       vector<double> ee(maxrad);
       vector<double> ec(fraction.size());

   // Loop over PSF image pixels and find the EE as a fn of r
       for (int nx=-maxrad; nx<maxrad; ++nx) {
           for (int ny=-maxrad; ny<=maxrad; ++ny) {

             // Position in PSF image in mm
               double x = DBL(nx) * px() + _baryX + _xOffset,
                      y = DBL(ny) * py() + _baryY + _yOffset;

             /* Set the pixel value to the value of the PSF
                value in the CCF rotated and translated to the
                required phi */
               int dist=round(sqrt(double(nx*nx+ny*ny)));

               if (dist < maxrad) 
                  ee[dist] += (*this)(x, y); 
           }
       }

   // Change from anuluar to cumulative
       for (int r=1; r<maxrad; ++r) 
                 ee[r] += ee[r-1];

   // DEBUG REMOVE
   //    for (int r=0; r<maxrad; ++r)
     //      cout<<"EE at: "<<r<<" "<<ee[r]<<endl;
   // DEBUG REMOVE

   // Find the radius which encloses the required set of fractions
   // Radii are in 1 arcsecond bins
       for (unsigned f=0; f<fraction.size(); ++f) {

           if (fraction[f] <= ee[0])
              ec[f] = fraction[f]/ee[0];
           else if (fraction[f] >=ee[maxrad-1])
              ec[f] = maxrad;
           else {

              for (int r=0; r<maxrad-1; ++r) {
                if (fraction[f] > ee[r] && fraction[f] < ee[r+1])
                  ec[f] = double(r)+1.0+ (fraction[f] - ee[r]) /
                                        (ee[r+1] - ee[r]);
              }
           }

           ec[f] *= px(); // Convert from arcsec to mm

       } // end of loop over fractions

       return ec;
}

vector<double> XmmPsf::enclosingCircle(vector<double> fraction) const
{
       vector<double> ec;

       for (unsigned f=0; f<fraction.size(); ++f) 
           ec.push_back(enclosingCircle(fraction[f]));

       return ec;
}

// Method: centreOffset - returns the offset of PSF centre in arcsecs
//                      - of a position theta (units of arcsecs)
double EbetaAnalyticPsf::centreOffset(double thetaAsec) const
{
   // Calculate the off-axis angle in arcminutes
    double theta = thetaAsec / 60.0;
    //double posang = this->posang() * -1.0;

    double radialOffset = _cenOffCoeff[0] + _cenOffCoeff[1]*theta +
			  _cenOffCoeff[2]*theta*theta +
			  _cenOffCoeff[3]*theta*theta*theta;

    return radialOffset;
} // EbetaAnalyticPsf::centreOffset

// Method: centreOffset - returns the offset of PSF centre in arcsecs
double EbetaAnalyticPsf::centreOffset() const
{
    return centreOffset(this->theta()*180.0*3600.0/M_PI);
} // EbetaAnalyticPsf::centreOffset()


double EbetaAnalyticPsf::enclosingCircle(double fraction) const
{
	if (fraction < 0. || fraction >= 1) {
		errstr<<"Given encircled energy fraction "<<fraction<<" is outside "
				"allowed interval [0, 1]"
			  <<Sas::warning("IllegalInputValue");
		return 0.;
	}

	//	see [3], Eq. (11)
        //      This needs updating for the gaussian bit
	return _param[0]*sqrt(pow(1.-fraction/_fac/_norm, -1./_alpha1) - 1.)*px();
} // EbetaAnalyticPsf::enclosingCircle()

/* Method appropriate only for the EllBeta model which adds the
   spokes and triangle and pentangle shape onto the base PSF */
//Image<float> EbetaAnalyticPsf::addSpokes(Image<float> &psfmapin) const
Image<float> EbetaAnalyticPsf::addSpokes(Image<float> &psfmap) const
{
// COMMENTS FROM original IDL code
// flat-topped triangular - distance between spokes = 2.0
// i.e. consider distance from spoke peak to trough centre as 1.0
// - u (half-width of flat top)
// - v (width of 1 slope bit from right edge of flat top to left edge of bottom)
// Nsp - No of spokes, rthresh - radius (pix) beyond which to start azim filtering
// changefrac - frac of image to be filtered (0[none]-1[all])
// secfrac - frac to go into secondary spoke component : 0.065
// M1frac - pentangle filtered
// M2frac - triangle filtered
// either** M1frac=0.13 or** M2frac=0.45
// rsmooth set to 2xrthresh - no smooth at centre - all smooth
//                            (v.v.light) at 2 x rthresh and beyond
// many constants read from the CCF lots (see below)

      // Set constants
        //cout<<" POS ANGLE: "<<this->posang()<<endl;
        //cout<<" NEW PHI: "<<this->phi()<<endl;
       /* Set the position angle which is used to rotate the spokes and
	  shape. The sign is switched to take into account that the
	  position angle increases anti-clockwise in image coords */
        double posang = this->posang() * -1.0; 

        double radian = 180.0 / M_PI;

      // These constants now to be read from the CCF
	double ophi=_ophi; // Phi to use for spoke pattern
	int Nsp=_nsp;   // Number of spokes
        double rthresh=_rthresh;  // 
        double changefrac=_changefrac;
        double M1frac=_m1frac; // Strength of the pentangular pattern
        double M2frac=_m2frac; // Strength of the triangular pattern
        double secfrac=_secfrac;
        double u=_u;
        double v=_v;
        double rd1=_rd1;       // The radius where the spoke begins
        double rd2=_rd2;       // The radius where the spoke strength peaks (")
        double rd3=_rd3;       // The radius where the spoke ends

        double h = (2 - (2*u) - (v))/((2*u) + v);
        double arc = M_TWO_PI / Nsp;
        double rophi=ophi/radian;
        double bit=1e-20;
        double intot=0.0; // Image sum before adding the spokes
        double outtot=0.0; // Image sum after adding the spokes
        //double f=0.0;
        bool mainspokes = true;
        //bool mainspokes = false;
        bool rsmooth = true;
        //bool rsmooth = false;   
        double dx=0.0, dy=0.0, ang=0.0;

       // Find image sum for final renorm
        for (unsigned ix=0; ix<psfmap.xExtent(); ++ix) {
          for (unsigned iy=0; iy<psfmap.yExtent(); ++iy) {
              intot += psfmap[ix][iy];   
          }
        }

        //cout<<"ELLIP: "<<_ellipity<<" TOT in: "<<intot<<" XX "<<intot/(1.0-_ellipity)<<endl;

             //psfmap[iy][ix]=psfmapin[ix][iy] * 625.0 * 625.007/514.957;//  M1
             //psfmap[iy][ix]=psfmapin[ix][iy] * 625.0 * 625.007/487.136; // M2

      // Find image sizes and centre
        unsigned nx = psfmap.xExtent();
        unsigned ny = psfmap.yExtent();
	int xc=nx/2+1;   // X centre of source  ! IS THIS RIGHT ??????
	int yc=ny/2+1;   // Y   "

      // Set temporary output image structure
        Image<float> im2(nx,ny);

      /* MOS1 structure hardwired - 5 'peaks' equi-angular at 62, 134, 
	 206, 278, 350 degrees clockwise of north in PA-corrected XY i
         sky space ('H'-space) */
	double M1phi = 62.0 / radian;
        double arcM1 = M_TWO_PI / 5.0;

      /* MOS2 structure hardwired - 3 'peaks' equi-angular at 50, 170, 
         290 degrees clockwise of north in PA-corrected XY sky space 
         ('H'-space) */
	double M2phi = 50.0 / radian;
	double arcM2 = M_TWO_PI / 3.0;

        
      /* -----------------------
	 Main spikes 
         ----------------------- */
       if (mainspokes) {

	for (unsigned ix=0; ix<nx; ++ix) {
	  for (unsigned iy=0; iy<ny; ++iy) {

	    dx= double(ix+1.0-xc);
            dy= double(iy+1.0-yc);

           /* Work on spoke if between an inner and outer radius */
            double r = sqrt( (dx*dx) + (dy*dy) );
	    //if ( r > rthresh && r<rd3) {
	    if ( r > rd1 && r<rd3) {

    	       ang = atan2(dx,(dy+bit)) - posang;
    	       if (ang < 0.0) {ang = M_TWO_PI + ang;}

    	       double ang2=(ang-rophi-(floor(ang/arc)*arc))*Nsp; // Main spokes
       	       if (ang2 < 0.0) {ang2 = M_TWO_PI + ang2;}
    	       double ang3=(ang-0.0-(floor(ang/arc)*arc))*Nsp; // Secdry spokes
    	       if (ang3 < 0.0) {ang3 = M_TWO_PI + ang3;}

             // Find the 'f' factor for the main spokes
               double fmain = spokeAngFactor(ang2,u,v,h);
               double fsec = spokeAngFactor(ang3,u,v,h);

             // The spokes have a radial dependence which starts at r1,
             // peaks at rd2 and reduces to zero at rd3
               double rfac = (1.0 - (r-rd2)/(rd3-rd2));
               if (r<rd2) rfac = (r-rd1) / (rd2-rd1);
               
             // Add the main and secondary spokes onto the original image
               //im2[ix][iy] = psfmap[ix][iy] * (1.0 + changefrac * rfac * (
                 //       ((1.0-secfrac) * fmain) + (secfrac * fsec) ) ); 
               im2[ix][iy] = psfmap[ix][iy] * 
                     ( (1.0 + changefrac * rfac * fmain) * (1.0-secfrac) +
                      (1.0 + changefrac * rfac * fsec) * secfrac );


             // DEBUG
             // if (r<10.0) { cout<<dx<<" "<<dy<<" "<<r<<" "<<rfac<<endl;}

            }
            else {
               im2[ix][iy]=psfmap[ix][iy];
            }
           // Find sum of elements
	    outtot += im2[ix][iy];
          } // iy loop
        } // ix loop
          //  cout<<"Before spokes total: "<<intot<<endl;
            // cout<<"After spokes total: "<<outtot<<endl;

      } else {  // For debugging
          for (unsigned ix=0; ix<nx; ++ix) 
            for (unsigned iy=0; iy<ny; ++iy) 
                im2[ix][iy]=psfmap[ix][iy];
      }

       /* Add in the shape for MOS PSFs */
        outtot=0.0;
        if (calServer.instrument()->id() != Xmm::EPN) {

          for (unsigned ix=0; ix<nx; ++ix) {
            for (unsigned iy=0; iy<ny; ++iy) {

               dx= double(ix+1.0-xc);
               dy= double(iy+1.0-yc);

               ang=atan2(dx,(dy+bit)) - posang;
               if (ang < 0.0) {ang = M_TWO_PI + ang;}

              // Apply shape filtering outside inner threshold
               if (sqrt( (dx*dx) + (dy*dy) ) > rthresh) {

                 if (calServer.instrument()->id() == Xmm::EMOS1) {

                // M1 - cos like pentangle function
                  double angM1=(ang-M1phi-(floor(ang/arcM1)*arcM1))*5.0;
                  if (angM1 < 0.0) {angM1 = M_TWO_PI + angM1;}
        	  im2[ix][iy]=im2[ix][iy] + (M1frac * cos(angM1) *im2[ix][iy]);

                 } else if (calServer.instrument()->id() == Xmm::EMOS2) {

                // M2 - cos like triangle function
                  double angM2=(ang-M2phi-(floor(ang/arcM2)*arcM2))*3.0;
                  if (angM2 < 0.0) {angM2 = M_TWO_PI + angM2;}
        	  im2[ix][iy]=im2[ix][iy] + (M2frac * cos(angM2) *im2[ix][iy]);

                 }

               } //else {
                  //im2(x,y)=im(x,y);
               //}
	    outtot += im2[ix][iy];
            }
          }
        }

      /* radius-smoothing image */
        Image<float> smim=im2;  // Temp array for smoothed image

        if (rsmooth) {

          double rsmthresh=2*rthresh;
          //double rsmthresh=2*rthresh / 1.1;  // PROBLEM !!!

          for (unsigned ix=1; ix<(nx-1); ++ix) {
            for (unsigned iy=1; iy<(ny-1); ++iy) {

               // Set the output pixel to the average of the surrounding
               // 9 pixels
                double pixout=0.0;
                for (unsigned aa=ix-1; aa<=ix+1; ++aa) 
                 for (unsigned bb=iy-1; bb<=iy+1; ++bb) 
                    pixout+=im2[aa][bb];

                smim[ix][iy]=pixout/9.0;
            }
          }

         // Add the smoothed image to the real image 
          outtot=0.0;
          for (unsigned ix=0; ix<nx; ++ix) {
            for (unsigned iy=0; iy<ny; ++iy) {

               dx= double(ix+1.0-xc);
               dy= double(iy+1.0-yc);

               double r = sqrt( (dx*dx) + (dy*dy) );
               double rfac = (1.0 - ( r / rsmthresh) );
               rfac = max(0.0,rfac);
               im2[ix][iy] = (rfac * im2[ix][iy]) + ((1.0-rfac)*smim[ix][iy]);

    	       outtot += im2[ix][iy];

            }
          }
        }

      /* Renormalise to make sure input has the same sum as input.
         Also, swap the axes because of the IDL issue */
  //        Image<float> outmap(psfmapin.xExtent(),psfmapin.yExtent());
          double finaltot=0.0;
          for (unsigned ix=0; ix<nx; ++ix) 
            for (unsigned iy=0; iy<ny; ++iy) {

               // outmap[iy][ix]=im2[ix][iy] * intot / outtot;
                im2[ix][iy]=im2[ix][iy] * intot / outtot;

                finaltot += im2[ix][iy];
            }

        //return outmap;  // Return the axes-swapped, renormalised image
        return im2;  // Return the renormalised image

} // EbetaAnalyticPsf:addSpokes

 /* Calculates the spoke factor for this angle */
double EbetaAnalyticPsf::spokeAngFactor(double ang2, double u, double v, double h)
const 
{
      double f = 0.0;

             /* flat-topped triangular */
              //flattop
               if (ang2 < M_TWO_PI*(u/2.0) || ang2 > M_TWO_PI*(1 - (u/2.0))) {
                  f=h;
               }
              //bottom
               if ( (ang2 > (M_TWO_PI*(u+v)/2.0)) &&
                       (ang2 < (M_TWO_PI*(1 - (u+v)/2.0))) ) {
                  f=-1.0;
               }
              // (down)
               if (ang2 >= M_TWO_PI*(u/2.0) &&
                       ang2 <= M_TWO_PI*((u+v)/2.0)) {
                  f=h - ( ((ang2-(M_PI*u/1.0))/(M_PI*v/1.0)) * (h+1) );
               }
              // slope (up)
               if (ang2 <= M_TWO_PI*(1 - (u/2.0)) &&
                       ang2 >= M_TWO_PI*(1 - ((u+v)/2.0))) {
                  f= h-( ((M_TWO_PI-(ang2+(M_PI*u/1.0))) /
                                           (M_PI*v/1.0)) * (h + 1));
               }

     return f;
}




EbetaAnalyticPsf::~EbetaAnalyticPsf()
{
} // EbetaAnalyticPsf:~EbetaAnalyticPsf()


//
//	---( SmallMapsBasedPsf )----------------------------------------------
//
SmallMapsBasedPsf::PsfMap::PsfMap(Array *a)
{
	Image<float> &i = *this;
	a>>i;
	_id=a->blockNumber();
	_E = a->attribute("ENERGY")->asReal();
	_theta = a->attribute("THETA")->asReal() * 60.; // arcmin->arcsec
} // PsfMap::PsfMap()


SmallMapsBasedPsf::SmallMapsBasedPsf(const Table *alignmentParamTable,
									 const CifEntry *cifEntry,
									 const Array *firstSmallMap)
	: XmmPsf(alignmentParamTable, firstSmallMap)
{
	_psfSet = cifEntry->open();
	_xExtent = firstSmallMap->size()[0];
	_yExtent = firstSmallMap->size()[1];
	unsigned fromArray = firstSmallMap->blockNumber(),
			 toArray = _psfSet->blocks();
	for (unsigned na=fromArray; na<toArray; ++na) {
		if (_psfSet->block(na)->type() == Block::TableT)
			//	skip tables
			continue;
		CcfConstituentArray ccfarr(_psfSet, na);
		const Array *a = ccfarr.block();
		if (INT(a->size()[0]) != _xExtent ||
			INT(a->size()[1]) != _yExtent)
			errstr<<"Size of array `"<<a->qualifiedName()<<"' is "
				  <<a->size()[0]<<'x'<<a->size()[1]<<" but expected "
				  <<_xExtent<<'x'<<_yExtent
				  <<Sas::error("CcfIntegrityError");

		double E = a->attribute("ENERGY")->asReal(),
			   theta = a->attribute("THETA")->asReal() * 60.; // arcmin->arcsec
		MIterOuter i = _idMap.find(E);
		if (i != _idMap.end()) {
			MIterInner j = i->second.find(theta);
			if (j != i->second.end())
				errstr<<"Have read PSF map for E/theta="<<E<<'/'<<theta
					  <<" already"<<Sas::error("CcfIntegrityError");
			i->second[theta] = na;
		} else {
			MapTypeInner innerMap;
			innerMap[theta] = na;
			_idMap[E] = innerMap;
		}
	}
	for (unsigned n=0; n<NUM_OF(_maps); _maps[n++]=0){};
	eValidity(
		ValidityInterval(_idMap.begin()->first, (--_idMap.end())->first));
	thetaValidity(
		ValidityInterval(0., (--_idMap.begin()->second.end())->first));
	phiValidity(ValidityInterval(0., M_TWO_PI));
} // SmallMapsBasedPsf::SmallMapsBasedPsf()


const Psf &SmallMapsBasedPsf::setParameters(double E,
											double theta,
											double phi)
{
	XmmPsf::setParameters(E, theta, phi);

	MIterOuter ior = _idMap.upper_bound(E);
	if (ior == _idMap.end() || ior == _idMap.begin())
		errstr<<"PSF energy can only be chosen in the interval ["
			  <<_idMap.begin()->first<<", "<<(--_idMap.end())->first
			  <<"] - requested energy is: "<<E
			  <<Sas::error("EnergyOutOfValidBounds");
	MIterOuter iol = ior; --iol;

	MIterInner iiur = ior->second.upper_bound(theta);
	if (iiur == ior->second.end() || iiur == ior->second.begin())
		errstr<<"PSF off-axis angle can only be chosen in the interval ["
			  <<ior->second.begin()->first<<", "<<(--ior->second.end())->first
			  <<"] - requested angle is: "<<theta
			  <<Sas::error("OffAxisAngleOutOfValidBounds");
	MIterInner iilr = iiur; --iilr;

	MIterInner iiul = iol->second.upper_bound(theta);
	assert(iiul != iol->second.begin() && iiul != iol->second.end());
	MIterInner iill = iiul; --iill;

	MIterInner	itervec[] = { iill, iilr, iiur, iiul };

	unsigned N = NUM_OF(itervec);
	for (unsigned n=0; n<N; ++n) {
		unsigned id = itervec[n]->second;
		if (!_maps[n])
			_maps[n] = new PsfMap(_psfSet->array(id));
		else
			if (_maps[n]->id() != id) {
				unsigned i = n+1;
				for (; i<N; ++i)
					if (_maps[i]->id() == id) {
						PsfMap *tmp = _maps[n];
						_maps[n] = _maps[i];
						_maps[i] = tmp;
						break;
					}
				if (i>=N) {
					delete _maps[n];
					_maps[n] = new PsfMap(_psfSet->array(id));
				}
			}
	}

	double E0 = _maps[0]->E(),
		   E1 = _maps[1]->E(),
		   theta0 = _maps[0]->theta(),
		   theta1 = _maps[2]->theta();

	_tE = (E-E0)/(E1-E0);
	_ut = (theta-theta0)/(theta1-theta0);


	return *this;
} // SmallMapsBasedPsf::setParameters()


double SmallMapsBasedPsf::pointValue(double x, double y) const
{
	int ix = INT(x+.5),
		iy = INT(y+.5);

// RDS - remove
//        cout<<setprecision(32)<<"IX IY: "<<ix<<" "<<iy<<" "<<x<<" "<<y<<endl;

	if (ix<0. || ix>=_xExtent ||
		iy<0. || iy>=_yExtent)
		return 0.;
//		errstr<<"Requested PSF at point ("<<ix<<','<<iy<<") which is out "
//				"of valid bounds [0, "<<_xExtent-1<<"] x [0, "<<_yExtent-1
//			  <<"]"<<Sas::error("positionOutOfBounds");


	PsfMap &psfLL = *_maps[0],
		   &psfLR = *_maps[1],
		   &psfUR = *_maps[2],
		   &psfUL = *_maps[3];

      // DEBUG //
	//cout<<"MAP[0]80,60: "<<iy<<" "<<ix<<" "<<psfLL[258][255]*1000.0<<" "<<psfLL[255][258]*1000.0;
//	return (1.-_tE)*(1.-_ut) * psfLL[ix][iy] +
//		   _tE*(1.-_ut) * psfLR[ix][iy] +
//		   _tE*_ut * psfUR[ix][iy] +
//		   (1.-_tE)*_ut * psfUL[ix][iy];

       /* The small maps have their X and Y axes flipped when being read in.
          Here we flip them back again. */
	return (1.-_tE)*(1.-_ut) * psfLL[iy][ix] +
  		   _tE*(1.-_ut) * psfLR[iy][ix] +
		   _tE*_ut * psfUR[iy][ix] +
		   (1.-_tE)*_ut * psfUL[iy][ix];

} // SmallMapsBasedPsf::pointValue()


SmallMapsBasedPsf::~SmallMapsBasedPsf()
{
	dataSetServer->close(_psfSet);
	for (unsigned n=0; n<NUM_OF(_maps); delete _maps[n++]){};
} // SmallMapsBasedPsf::~SmallMapsBasedPsf()


//
//	---( PsfDataServer )--------------------------------------------------
//
PsfDataServer::PsfDataServer(Psf *psf,
							 const Table *betaPsfTab,
							 CalAtomCollection<PsfDataServer> *parent)
	: CalChild<CalAtomCollection<PsfDataServer> >(parent), _psf(psf)
{
	_hasBetaPsf = betaPsfTab;
	if (!_hasBetaPsf) {
		if (calServer.instrument()->behindTelescope() != Xmm::XRT3)
		errstr<<"CCF::XRT?_PSF does not contain a BETAPSF table - "
				"PsfDataServer::dispersionFigureDistribution() will not work"
			  <<Sas::warning("NoSuchTable");
		return;
	}

	vector<double>	norm, center, width;
	vector<string>	type;
	CalOalUtility::fillVector(betaPsfTab, "TYPE", type);
	CalOalUtility::fillVector(betaPsfTab, "NORM", norm);
	CalOalUtility::fillVector(betaPsfTab, "CENTER", center);	// [rad]
	CalOalUtility::fillVector(betaPsfTab, "WIDTH", width);		// [rad]

	unsigned N = type.size();
	for (unsigned n=0; n<N; n++) {
		unsigned i = type[n][0] == 'G' ? 0 : 1;
		_norm[i].push_back(norm[n]);
		_center[i].push_back(center[n]);
		_width[i].push_back(width[n]);
	}
} // PsfDataServer::PsfDataServer()


void PsfDataServer::dump(ostream &)
{
	errstr<<"Sorry dumping PsfDataServer not yet implemented"
		  <<Sas::warning("NotYetImplemented");
} // PsfDataServer::dump()


Psf::ValidityInterval PsfDataServer::eValidity() const
{
	assert(_psf);
	return _psf->eValidity();
} // PsfDataServer::eValidity()


Psf::ValidityInterval PsfDataServer::thetaValidity() const
{
	assert(_psf);
	return _psf->thetaValidity();
} // PsfDataServer::thetaValidity()


Psf::ValidityInterval PsfDataServer::phiValidity() const
{
	assert(_psf);
	return _psf->phiValidity();
} // PsfDataServer::phiValidity()


Psf &PsfDataServer::psf(double E, double theta, double phi)
{
	_psf->setParameters(E, theta, phi);	

	return *_psf;	
} // PsfDataServer::psf()


Spectrum PsfDataServer::dispersionFigureDistribution(
											  double nBetaHWHM,
											  double betaBinWidth,
											  double chromaticMagnification,
											  double inChannelOffset,
											  bool analytical)
{
	if (!_hasBetaPsf)
		errstr<<"CCF does not contain BETAPSF table - cannot continue"
			  <<Sas::error("NoSuchTable");

	if (!analytical)
		errstr<<"Numerical integration of intrinsic PSF in cross-dispersion "
				"direction not implemented yet - reverting to analytical "
				"model"<<Sas::warning("NotImplementedYet");

	//	copy function parameters
	vector<double> norm[2] = { _norm[0], _norm[1] },
				   center[2] = { _center[0], _center[1] },
				   width[2] = { _width[0], _width[1] };
	//	index 0: Gaussian
	//		  1: Lorentzian

	unsigned N[2] = { norm[0].size(), norm[1].size() };
	for (unsigned n=0; n<N[0]; ++n) {
		norm[0][n] *= M_SQRT_TWO_PI * width[0][n];
	}

	//	total normalizations
	double tot[2];
	for (unsigned i=0; i<2; ++i) {
		tot[i] = 0.;
		for (unsigned n=0; n<N[i]; tot[i] += norm[i][n++]){};
	}
	double totNorm = tot[0] + tot[1];

	//	calculate first moment and renormalize individual amplitudes
	double firstMoment[2];	//	in beta space
	for (unsigned i=0; i<2; ++i) {
		firstMoment[i] = 0.;
		for (unsigned n=0; n<N[i]; ++n) {
			norm[i][n] /= totNorm;
			firstMoment[i] += norm[i][n]*center[i][n];
		}
	}

	double firstMom =
				(tot[0]*firstMoment[0]+tot[1]*firstMoment[1])/(tot[0]+tot[1]);

	//	apply chromatic magnification correction to centers and sigmas
	double maxWidth = 0;
	for (unsigned i=0; i<2; ++i)
		for (unsigned n=0; n<N[i]; ++n) {
			center[i][n] = chromaticMagnification*(center[i][n]-firstMom)+
																	firstMom;
			width[i][n] *= chromaticMagnification;
			if (width[i][n] > maxWidth)
				maxWidth = width[i][n];
		}

	//	go to channel space - also adjust the centers such that the
	//	first moment lies at <inChannelOffset> from the array center
	unsigned nBetaBins = 1 + 2 * UNT(maxWidth/betaBinWidth * nBetaHWHM + .5);
	double betaBins = DBL(nBetaBins),
		   betaSpan = betaBins * betaBinWidth,
		   chanPerRad = betaBins/betaSpan,
		   firstMomChan = firstMom * chanPerRad,
		   deltaCenter = inChannelOffset - firstMomChan;

	for (unsigned i=0; i<2; ++i)
		for (unsigned n=0; n<N[i]; ++n) {
			center[i][n] = center[i][n] * chanPerRad + deltaCenter;
			width[i][n] *= chanPerRad;
	}

	//
	//	fill the spectrum
	//
	Spectrum spec(nBetaBins, 0.);
	int iBetaBins = nBetaBins/2;
	for (int ii=1; ii>=-1; ii-=2)
		for (int i=0; i<=iBetaBins; ++i) {
			int indx = ii*i;
			spec[indx+iBetaBins] = 0.;
			//	add up Gaussians
			for (unsigned n=0; n<N[0]; ++n) {
				double oneOversq2sigma = 1./(M_SQRT_2*width[0][n]),
					   argPlus = DBL(indx)+.5-center[0][n],
					   argMinus = DBL(indx)-.5-center[0][n];
				spec[indx+iBetaBins] += .5*norm[0][n] *
							fabs(Slatec::erf(oneOversq2sigma * argPlus) -
								 Slatec::erf(oneOversq2sigma * argMinus));
			}
			//	add up Lorentzians
			for (unsigned n=0; n<N[1]; ++n) {
				double normOverPi = norm[1][n]/M_PI,
					   argPlus = DBL(indx)+.5-center[1][n],
					   argMinus = DBL(indx)-.5-center[1][n];
				spec[indx+iBetaBins] += normOverPi *
											fabs(atan2(argPlus, width[1][n]) -
												 atan2(argMinus, width[1][n]));
			}
		}

	//	set first moment
	spec.firstMoment(iBetaBins + inChannelOffset);

	SAS_DEBUG({
		unsigned NN = spec.size();
		double ntot = 0.;
		for (unsigned n=0; n<NN; ntot+=spec[n++]);
		cout<<NN<<' '<<ntot<<' '<<spec.firstMoment()<<endl;
	})

	return spec;
} // PsfDataServer::dispersionFigureDistribution()


PsfDataServer::~PsfDataServer()
{
	delete _psf;
} // PsfDataServer::~PsfDataServer()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<PsfDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	PsfDataServer *null = 0;
	_collection = new vector<PsfDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::XPSF);
	DataSet *set = e->open();
	const Table *betaPsfTab = 0;
	if (set->hasBlock("BETAPSF")) {
		CcfConstituentTable ccftab(set, "BETAPSF");
		betaPsfTab = ccftab.block();
	}

	//	check if there is a KING_PARAMS block
	if (accuracy == EXTENDED && !set->hasBlock("KING_PARAMS")) {
		errstr<<"The CCF constituent `"<<set->qualifiedName()<<"' does not "
				"contain a table with King-PSF parameters - reverting to "
				"HIGH accuracy level."
			  <<Sas::warning("MissingData");
		calServer.state()->accuracyLevel()->set(HIGH);
		accuracy = HIGH;
		_accuracy = HIGH;
	}

	//	check if there is an ELLBETA_PARAMS block
	if (accuracy == ELLBETA && !set->hasBlock("ELLBETA_PARAMS")) {
	   if (set->hasBlock("KING_PARAMS")) 
              accuracy = EXTENDED;
           else
              accuracy = HIGH;
           
 	   errstr<<"The CCF constituent `"<<set->qualifiedName()<<"' does not "
				"contain a table with Elliptical King function parameters - reverting to " <<accuracy<<" accuracy level."
		  <<Sas::warning("MissingData");
	   calServer.state()->accuracyLevel()->set(accuracy);
	   _accuracy = accuracy;
	}

        //      check if there is a SLEWPSF_PARAMS block
        if (accuracy == SLEW && !set->hasBlock("SLEWPSF_PARAMS")) {
              accuracy = ELLBETA;

           errstr<<"The CCF constituent `"<<set->qualifiedName()<<"' does not "
                                "contain a table with slew psf parameters - reverting to " <<accuracy<<" accuracy level."
                  <<Sas::warning("MissingData");
           calServer.state()->accuracyLevel()->set(accuracy);
           _accuracy = accuracy;
        }


	CcfConstituentTable ccftab0(set, "ALIGNMENT_PARAMS");
	switch (accuracy) {
		case LOW: {
			//	single Gauss model
			CcfConstituentTable ccftab1(set, "SINGLE_GAUSS_PARAMS");
			(*_collection)[0] = new PsfDataServer(
									new CslAnalyticPsfLow(
										ccftab0.block(),
										ccftab1.block()),
									betaPsfTab, this);
			break;
		}
		case MEDIUM: {
			//	SciSim PSF maps
			unsigned n, nB = set->blocks();
			for (n=1; n<nB; ++n) {
				if (set->block(n)->type() == Block::ArrayT)
					break;
			}
			if (n>=nB)
				errstr<<"Now PSF maps found in '"<<set->qualifiedName()
					  <<"'"<<Sas::error("CcfIntegrityError");
			CcfConstituentArray ccfarr1(set, n);
			(*_collection)[0] = new PsfDataServer(
									new SmallMapsBasedPsf(
										ccftab0.block(), e,
										ccfarr1.block()), betaPsfTab, this);
			break;
		}

		case EXTENDED: {
			//	King model
			CcfConstituentTable ccftab1(set, "KING_PARAMS");
			(*_collection)[0] = new PsfDataServer(new CnrAnalyticPsf(
													ccftab0.block(),
													ccftab1.block()),
												  betaPsfTab, this);
			break;
		}

		case ELLBETA: {
			//	EBG model

			CcfConstituentTable ccftab1(set, "ELLBETA_PARAMS");

			(*_collection)[0] = new PsfDataServer(
                                                 new EbetaAnalyticPsf(
						      ccftab0.block(),
						      ccftab1.block()),
						      betaPsfTab, this);
			break;
		}

		case SLEW: {
			//	EBG model

			CcfConstituentTable ccftab1(set, "SLEWPSF_PARAMS");

			(*_collection)[0] = new PsfDataServer(
                                                 new EbetaAnalyticPsf(
						      ccftab0.block(),
						      ccftab1.block()),
						      betaPsfTab, this);
			break;
		}

		case HIGH: {
			//	three-Gaussian model
			CcfConstituentTable ccftab1(set, "TRIPLE_GAUSS_PARAMS");
			(*_collection)[0] = new PsfDataServer(
									new CslAnalyticPsfMedium(
										ccftab0.block(),
										ccftab1.block()),
									betaPsfTab, this);
			break;
		}
		default:
			break;
	}

	dataSetServer->close(set);
} // CalAtomCollection<PsfDataServer>::CalAtomCollection()



/* AI gcc4 porting
 * too few template-parameter-lists -> template<>
 */
template<>
PsfDataServer *CalAtomCollection<PsfDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<PsfDataServer>::select()
