// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2020 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	RmfEpochServer.cc --- implementation of class to return data from
//			     the RMF epoch CCF
//
//	DESCRIPTION:
//
//              This atom returns the epoch of the canned response matrice
//		corresponding with a given date
//
//	CHANGE HISTORY:
//		RDS, 2023-08-10: 1.0.0: - original
//
//	AUTHOR:
//		R. Saxton, 2023-08-10
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//

#include "RmfEpochServer.h"
#include "CalOalUtility.h"
#include "CalServer.h"
#include "STime.h"

#include <algorithm>
#include <cmath>

using namespace CalOalUtility;
using namespace std;

//
//	----------------------( mode parameters )-----------------------------
//

//  Define the RmfEpochServer class
//
RmfEpochServer::RmfEpochServer( 
        std::vector<Pair> epochdate,
        CalAtomCollection<RmfEpochServer> *parent)
        : CalChild<CalAtomCollection<RmfEpochServer> >(parent), _epochdate(epochdate)
{
}

// Return the epoch for this instrument and date
const short RmfEpochServer::getRmfEpoch(const string &indate)
{
  // Check that the input date string is valid. If so, convert to an STime object.
  try {
         STime deldate(indate);
  } catch(SasException::Error) {
         errstr<<"Input date: "<<indate<<" is not valid"
                   <<Sas::error("InvalidDateTime");
  }
  STime sdate(indate);

  // Return the epoch for this date
  // Dates in the vector are the initial date for which that epoch is valid
  for (unsigned i = 1; i<_epochdate.size(); ++i) {
      //cout<<i<<" "<<_epochdate[i].second<<endl;
      if (sdate < STime(_epochdate[i].second)) {
         //cout<<"Found it: "<<_epochdate[i].second;
         return _epochdate[i-1].first;
      }
  }

  // Must be the last epoch if date is greater than the last one
  return _epochdate[_epochdate.size()-1].first;
}

template<>
CalAtomCollection<RmfEpochServer>::CalAtomCollection(
	                                     const XmmInstrument *instrument,
					     const Ccf *ccf,
					     Xmm::Filter filter,
					     AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nModes		= instrument->nModes();
        _nCCDs		= instrument->nCcds();

	RmfEpochServer *null = 0;
	_collection = new vector<RmfEpochServer *>(1, null);

	CifEntry *e = ccf->entry(instrument, Ccf::RmfEpoch);
	DataSet *dataSet = e->open();

      // Open the rmf epoch extension
        Table *table = dataSet->table(e->queryBlockName("RMFEPOCH"));

      // Read epch and start_date into vectors
        vector<string> startDate;
        vector<short> epoch;
        ColumnData *insCol = table->column("INSTRUME")->data();
        CalOalUtility::fillVector(table, "EPOCH", epoch);
        CalOalUtility::fillVector(table, "START", startDate);
        
        // Get current instrument as string
        string instrumentName(Xmm::toString(instrument->id()));

      // Produce a vector of pairs of epoch and start date for
      // the requested instrument
        std::vector<Pair> m;
   
        for (unsigned n=0; n<table->rows(); ++n) {
            string instId(uppercase(stringColumnRow(insCol, n)));
            if (instId == instrumentName) {
               m.push_back(make_pair(epoch[n], startDate[n]));
            }
        }

      // Create the collection - just one
        (*_collection)[0] = new RmfEpochServer(m, this);

	dataSetServer->close(dataSet);
} // CalAtomCollection<RmfEpochServer>::CalAtomCollection


/* AI gcc4 porting
 *  too few template-parameter-lists => template<>
 */
template<>
RmfEpochServer *CalAtomCollection<RmfEpochServer>::select(
					Ccd,
					Node,
					unsigned) const
{
        if (!(*_collection)[0])
           errstr<<"The CCF does not contain RmfEpoch data "
                          <<Sas::error("NoData");

	return (*_collection)[0];
} // CalAtomCollection<RmfEpochServer>::select
